#!
#
# Linux Live Response script: start with static bash from trusted source
#
# Collect volatile data from a linux system.
# Designed to work with only statically linked binaries
# from a mounted external media.
#
# It is assumed that all binaries are in the same path as "bash" is.
#
# No files will be touched or read directly.
# Exceptions: /etc/passwd, utmp, wtmp (in the user mappings & login section)
#
# It is assumed that a disk dump is made directly afterwards to
# preserve this information.
#
# $Id: llr.sh,v 1.42 2008/08/08 08:58:21 enno Exp $
#
# Version 0.4.1 08/2008 Enno Ewers
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA

VERSION="Version 0.4 \$Revision: 1.42 $"

# Paranoia - use builtins
enable cd break set unset true false test [ exit pwd shift shopt trap

#
# Fail if executed with system bash
# (This is not foolproof!)
#
if [ "${BASH%/*}" = "/bin" -o "${BASH%/*}" = "/usr/bin" -o \
     "${BASH%/*}" = "/sbin" -o "${BASH%/*}" = "/usr/sbin" -o \
     "${BASH%/*}" = "/usr/local/bin" -o "${BASH%/*}" = "/usr/local/sbin" ]; then
    echo "Please start script with static shell, e.g.:"
    echo "# <path-to-static-bins>/bash $0"
    exit 1
fi

#
# Tags
#
SECTION="===Section> "
SUBSEC="===Subsec> "
INFO="===Info> "
STARTCMD="===Cmd> "
EXITSTATUS="---Status> "

#
# Script options
#
declare -a allowed_options
declare -a options

allowed_options=(procmem kmem kmemlast checksum date pcat)
options=(checksum date procmem kmem)

#
# Function declarations
#
error() {
    echo "Error: $@" 1>&2
    echo "Error: $@"
    exit 1
}

usage() {
    echo "Usage: bash ./${0##*/} <host> <port>" 1>&2
    echo "   or: bash ./${0##*/} <outputfilename>" 1>&2
    exit 1
}

inf() {
    echo "I> ${CURSEC}: $*" 1>&2
    echo "${SUBSEC}$*"
}

log() {
    echo "$*"
}

loginfo() {
    echo "${INFO}$*"
}

section() {
    CURSEC="$*"
    echo "${SECTION}$*"
}

run() {
    BIN="$1"
    shift
    log "${STARTCMD}${BIN}" "$@"
    loginfo "full path: ${BINDIR}/${BIN}"
    if option checksum; then
	loginfo "sha256sum: "`"${BINDIR}/sha256sum" -b "${BINDIR}/${BIN}" | "${BINDIR}/cut" -d' ' -f1`
    fi
    if option date; then
	loginfo "date: "`"${BINDIR}/date"`
    fi
    { "${BINDIR}/${BIN}" "$@"; } 2>&1 | "${BINDIR}/tr" '\0' "${NULCHAR}"
    E=$PIPESTATUS
    log ""
    log "${EXITSTATUS}$E"
    if [ 0 -ne "$E" ]; then
	echo "==> Error running: ${BIN}" "$@" "(Exit code: $E)" 1>&2
    fi
}

encode_run() {
    FILENAME="$1"
    shift
    BIN="$1"
    shift
    log "${STARTCMD}${BIN}" "$@"
    loginfo "full path: ${BINDIR}/${BIN}"
    if option checksum; then
	loginfo "sha256sum: "`"${BINDIR}/sha256sum" -b "${BINDIR}/${BIN}" | "${BINDIR}/cut" -d' ' -f1`
    fi
    if option date; then
	loginfo "date: "`"${BINDIR}/date"`
    fi
    { stderrlog=`("${BINDIR}/${BIN}" "$@" | logenc "${FILENAME}"; exit ${PIPESTATUS}; ) 3>&1 1>&2 2>&3`; } 3>&1 1>&2 2>&3
    E=$?
    log "${stderrlog}"
    log "${EXITSTATUS}$E"
    if [ 0 -ne "$E" ]; then
	echo "==> Error running: ${BIN}" "$@" "(Exit code: $E)" 1>&2
    fi
}

logenc() {
    "${BINDIR}/uuencode" "$1"
}

toggleoption() {
    for sopt in $1; do
	for opt in "${allowed_options[@]}"; do
	   if [ "x$sopt" = "x$opt" ]; then
		# valid option
		for xopt in "${options[@]}"; do
		    if [ "x$sopt" = "x$xopt" ]; then
			# set - unset it
			options=(${options[@]%$sopt})
			echo "=> Removing $sopt"
			break 2;
		    fi
		done
		# not set - set it
		options+=($sopt)
		echo "=> Adding $sopt"
	   fi	
	done
    done
}

option() {
    ret=1;
    for opt in "${options[@]}"; do
	if [ "x$1" = "x$opt" ]; then
	    # valid option
	    ret=0;
	    break;
	fi
    done
    return $ret;
}



##############################
# Main part
##############################

SCRIPT="$0"
shopt -s extglob
SCRIPTDIR="${0%%*([^/])}"
SCRIPTNAME="${0##*/}"
BASHDIR="${BASH%%/*([^/])}"
shopt -u extglob
#BINDIR="${PWD}/${SCRIPTDIR}${STATICPATH}"
BINDIR="${BASHDIR}"
PATH=${BINDIR}

NULCHAR='\n'

TARGET=
REMOTE=
PORT=

#
# ARGV
#
if [ $# -eq 1 ]; then
    TARGET=$1
elif [ $# -eq 2 ]; then
    REMOTE=$1
    PORT=$2
    TARGET="/dev/tcp/${REMOTE}/${PORT}"
else
    usage
fi

if [ ! -x "${BINDIR}/bash" ]; then
	echo "Cannot find '${BINDIR}/bash'" 1>&2
	echo "Start script from its local directory (.../bash ./${SCRIPTNAME})" 1>&2
	exit 1
fi

ID=`"${BINDIR}/id" -u`

if [ "${ID}" -ne "0" ]; then
	echo "WARNING: Not running as root (ID=${ID})" 1>&2
	echo "         Most data acquisition will _fail_. " 1>&2
	echo "" 1>&2
	echo -n "Enter the letters \"ok\" to continue> " 1>&2
    	read dummy
	if [ "$dummy" != "ok" ]; then
		echo "" 1>&2
		echo "Aborting. Please restart as root." 1>&2
		exit 1
	fi
fi

echo "=== Start of llr.sh ${VERSION}" 1>&2
echo "" 1>&2
echo "Available options:" 1>&2
echo "  checksum  Checksum every external command with 'sha256sum -b' (which is" 1>&2
echo "            also external) before running it." 1>&2
echo "  date      Run 'date' before every command to log exact start times. If" 1>&2
echo "            not set, only start and end times of the script will be logged." 1>&2
echo "  kmem      Dump kernel memory from /dev/kmem (may not work, may only dump" 1>&2
echo "            the first Mbyte, may in rare cases hang the system)" 1>&2
echo "  kmemlast  The kernel memory dump may hang in rare cases, until now only"
echo "            seen on XEN kernels. With this option, it is executed last. Use"
echo "            with XEN kernels."
echo "  procmem   Dump process memory. Takes a long time. May hang in rare cases" 1>&2
echo "            if executed under X11." 1>&2
echo "  pcat      Use pcat (from The Coroners Toolkit) instead of pd (process dumper" 1>&2
echo "            from trapkit.de)" 1>&2
echo "" 1>&2
while true; do
    echo "Current options: >${options[*]}<" 1>&2
    echo "" 1>&2
    echo -n "Enter option to toggle or press enter to continue> " 1>&2
    read dummy
    if [ -z "$dummy" ]; then
	break;
    fi
    toggleoption "$dummy"
done
echo "" 1>&2

echo "=== Start of llr.sh ${VERSION}" 1>&2
echo "===       CMD: $0" "$@" 1>&2
echo "===      BASH: ${BASH}" 1>&2
echo "===      PATH: ${BINDIR}" 1>&2
echo "=== SCRIPTDIR: ${SCRIPTDIR}" 1>&2
echo "===   OPTIONS: ${options[@]}" 1>&2

if [ -n "${REMOTE}" ]; then 
    echo "===    OUTPUT: remote ${TARGET}" 1>&2
else
    echo "===    OUTPUT: local file '${TARGET}'" 1>&2
fi
echo "" 1>&2

if [ -n "${REMOTE}" ]; then 
    echo "Start TCP listener on target host $REMOTE port $PORT:" 1>&2
    echo "e.g.: # nc -l [-p] $PORT | tee logfile.out" 1>&2
else
    echo "Will write to logfile '${TARGET}'" 1>&2
fi
echo -n "Press Enter to start> "
read dummy
echo "" 1>&2

#
# audit script: kernel memory dump 
#
audit_dump_mem() {
    # this will only log the first 1 Meg on some 2.6 kernels depending
    # on kernel configuration
    # check for "devmem_is_allowed" in /proc/kallsyms
    # this may hang the system under some circumstances
    inf "Memory dump allowed? (devmem_is_allowed should appear when yes)"
    run grep devmem /proc/kallsyms
    
    inf "Base memory dump"
    encode_run dev-mem.dump dd if=/dev/mem
}

#
# audit script: main
#

{
    loginfo "llr.sh: ${VERSION}"
    loginfo "Command invocation: $0" "$@"
    loginfo "bash: ${BASH}"
    loginfo "PATH: ${BINDIR}"
    loginfo "SCRIPTDIR: ${SCRIPTDIR}"
    loginfo "Options: ${options[@]}"
    loginfo "Output: ${TARGET}"

    section Date/Time
    inf "Script started"
    run date

    section Selfcheck
    inf SHA256sums
    
    if [ -f "${SCRIPTDIR}/sha256sums.txt" ]; then
	run sha256sum "$0" "${SCRIPTDIR}/sha256sums.txt" "${BINDIR}"/*
	run sha256sum -w -c "${SCRIPTDIR}/sha256sums.txt"
    else
	run sha256sum "$0" "${BINDIR}"/*
    fi

#
# Basics / Kernel
#

    section Basics

    inf "Hostname"
    run hostname

    inf "System date and uptime"
    run date
    run uptime

    section Kernel

    inf "Kernel version"
    run uname -a
    run cat /proc/version

    section Basics

    inf "Loadavg"
    run cat /proc/loadavg

    inf "Script environment"
    run printenv
    run id

#
# Early stuff (most volatile stuff)
#
    section Kernel

    inf "Dmesg output"
    run dmesg

    inf "Kernel meminfo"
    run cat /proc/meminfo
    run cat /proc/vmstat

    if option kmemlast; then
	true;
    else
	if option kmem; then
	    audit_dump_mem
	fi
    fi

#
# FS
#
    section Filesystems

    inf "Mounted filesystems and storage"
    run cat /proc/mounts
    run cat /proc/partitions
    run cat /proc/swaps
    run df -h
    run df -k

#
# processes
#

    section Processes

    inf "Running processes overview"
    run ps -efl

#
# network
#

    section Network

    inf "Network interfaces"
    run ifconfig -a
    run ifconfig -s

    ## (see sysctl -A)
    #inf "Network interface parameters from /proc"
    #run cat /proc/sys/net/ipv4/ip_forward
    #for nif in /proc/sys/net/*/conf/*; do
    #	run grep -r . "${nif}"
    #done

    inf "Routing and arp tables"
    run netstat -rn
    run arp -n
    run cat /proc/net/arp

    inf "Running network connections"
    run netstat -apn --inet
    # ("-Di" does not exist since lsof is compiled without HASDCACHE)
    # run lsof -P -n -i -Di
    run lsof -P -n -i -V

    #inf "RPC services"
    #run rpcinfo -p


#
# all sockets
#
    inf "Complete socket info"
    run netstat -anp
    run lsof -n -P -V

#
# kernel
#
    section Kernel 

    inf "Kernel boot parameters"
    run cat /proc/cmdline

    inf "Kernel modules"
    run cat /proc/modules

    inf "Sysctl values (incl. network parameters)"
    run sysctl -A

    inf "Kernel symbols"
    if [ -e /proc/ksyms ]; then
	run cat /proc/ksyms
    fi
    if [ -e /proc/kallsyms ]; then
	run cat /proc/kallsyms
    fi

#
# hardware
#
    section Hardware

    inf "Hardware info and devices"
    run cat /proc/cpuinfo
    run cat /proc/interrupts
    run cat /proc/devices
    run cat /proc/ioports

#
# Stuff that touches the filesystem
#
    section Userinfo

    inf "User mappings and logins (reads /etc/passwd, utmp and wtmp)"
    run cat /etc/passwd
    run w
    run who
    run last

#
# Process memory (long running)
#
   section Processes
    inf "Running processes by /proc (some errors are expected)"
    for pid in /proc/[0-9]*; do
	run ls -nN  "${pid}"
	for file in "${pid}/"{cmdline,maps,status,stat,statm}; do
	    run cat "${file}"
	done
	encode_run "${pid}-environ.raw" cat "${pid}/environ"
	run ls -n "${pid}"/fd
    done

    if option procmem; then
	# pcat/pd will fail for processes that are already ptraced.
	# pcat/pd will fail on kernel processes.
	inf "Process memory dump"
	for pid in /proc/[0-9]*; do
	npid="${pid##*/}"
	if [ "${npid}" -ne "$$" ]; then
	    if option pcat; then
		# pcat
		encode_run "process-memory-${npid}.dump" pcat "${npid}"
	    else
		# pd
		encode_run "process-memory-pid-${npid}.dump" pd_v1.1_lnx -p "${npid}"
	    fi
	fi
	done
    fi

    section Date/Time
    inf "END of main part - start of additional tests (may hang)"
    run date

    section Additional Tests

    inf "Devices (may hang)"

    run ls -lR /dev
    
    run lspci -v

    for file in /proc/bus/*/devices; do
	run cat "${file}"
    done

    if option kmemlast; then
	if option kmem; then
	    audit_dump_mem
	fi
    fi

    section Date/Time
    inf "END of script"
    run date

} >"${TARGET}"

echo "Done." 1>&2

if [ -n "${REMOTE}" ]; then 
    echo "Do not forget to take the checksum on the remote host!" 1>&2
    echo "e.g.: # sha256sum logfile.log > logfile.log.sha256sum" 1>&2
else
    echo "Writing sha256sum to: ${TARGET}.sha256sum" 1>&2
    "${BINDIR}/sha256sum" "${TARGET}" | "${BINDIR}/tee" "${TARGET}.sha256sum" 1>&2
    echo "" 1>&2
fi
